(define-module (evms)
  :use-module (ice-9 optargs)
  :use-module (ice-9 rdelim)
  :use-module (ice-9 popen)
  :use-module (ice-9 pretty-print)
  :use-module (srfi srfi-1)
  :use-module (srfi srfi-2)
  :use-module (srfi srfi-11)
  :use-module (srfi srfi-13)
  :export (engine))

(load-extension "libguile-evms.so" "scm_init_evms_module")

(define changed-root #f)

(define well-known-mountpoints
  '(("/" . "relatime")
    ("/home" . "nosuid,relatime")
    ("/usr" . "nodev,relatime")
    ("/var" . "nosuid,relatime")
    ("/boot" . "nodev,nosuid,noexec,relatime")
    ("/srv" . "nosuid,nodev,relatime")
    ("/opt" . "nosuid,nodev,relatime")
    ("/usr/local" . "nodev,relatime")
    ("/var/lib/vz" . "nodev,nosuid,noexec,relatime")
    ("/var/ftp" . "nodev,nosuid,relatime")
    ("/var/www" . "nodev,relatime")
    ("/var/cache/squid" . "nodev,noexec,nosuid,relatime")
    ("/boot/efi" . #f)
    ("/mnt/disk" . #f)))

(define bare-mountpoints '("/boot"))

;;------------------------------------------------------------------
;;          convenience utils
;;------------------------------------------------------------------
(define (obj->handle obj)
  (and obj
       (procedure? obj)
       (procedure-property obj 'handle)))

(define (any->handle obj)
  (or (obj->handle obj) obj))

(define (any->object thing)
  (if (number? thing)
      (and (not (eqv? thing 0))
           (object thing))
      thing))

(define (mapkey obj keys)
  (map (lambda (x) (cons (car x) ((cdr x) obj))) keys))

(define (isswap? vol)
  (and-let*
   ((fsim (assv-ref (vol) 'fsim)))
   (string=? "SWAPFS" (assv-ref (fsim) 'name))))

(define (isntfs? vol)
  (and-let*
   ((fsim (assv-ref (vol) 'fsim)))
   (string=? "NTFS" (assv-ref (fsim) 'name))))

(define (isfat? vol)
  (and-let*
   ((fsim (assv-ref (vol) 'fsim)))
   (string-prefix? "FAT" (assv-ref (fsim) 'name))))

(define (islvm? vol)
  (and-let*
   ((plugin (assv-ref ((assv-ref (vol) 'object)) 'plugin)))
   (string=? "LVM2" (assv-ref (plugin) 'name))))

(define (mntpoint? vol)
  (object-option-value vol 'mntpoint))

(define (mntpoint-nonempty? obj val)
  (and val (not (string=? val ""))))

(define (mntpoint-sane-pathname? obj val)
  (or (string=? val "/")
      (string=? val "swap")
      (and (let ((components (string-split val #\ )))
             (and (< (length components) 2))) ; no spaces please
           (let ((components (string-split val #\/)))
             (and (< 1 (length components)) ; at least one /
                  (equal? (car components) "") ; and leading one
                  (not (member "." (cdr components)))
                  (not (member ".." (cdr components)))
                  (not (member "" (cdr components)))))
           (not (member val '("/sys" "/proc" "/dev" "/dev/pts" "/lost+found"))))
      (throw 'swig-system-error
             "Assign mountpoint" "Invalid mountpoint")))

(define (mntpoint-valid-for-storage? obj val)
  (cond ((and (member val bare-mountpoints)
              (islvm? obj))
         (throw 'swig-system-error
                "Assign mountpoint" "Invalid mountpoint: should not be resided on LVM"))
        (#t #t)))

(define (mntpoint-valid-for-fstype? obj val)
  (cond
   ((isswap? obj)
    (or (string=? val "swap")
        (throw 'swig-system-error
               "Assign mountpoint" "Invalid mountpoint for SWAPFS")))

   ((string=? val "swap")
    (or (isswap? obj)
        (throw 'swig-system-error
               "Assign mountpoint" "Invalid mountpoint: not a SWAPFS")))

   ((member (assv-ref ((assv-ref (obj) 'fsim)) 'name) '("FAT16" "FAT32" "NTFS"))
    (or (string-prefix? "/mnt/" val)
        (string=? val "/boot/efi")
        (throw 'swig-system-error
               "Assign mountpoint" "Invalid mountpoint: should start with /mnt")))
   (#t #t)))

(define (mntpoint-unique? obj val)
  (or (and (mntpoint? obj) (string=? val (mntpoint? obj))) ; already has that value FIXME combo/ex
      (not (and-let*
            ((oldvol (find (lambda (v) (string=? val (mntpoint? v)))
                           (assigned-volumes))))
            (clear-object-options! oldvol)
            #f))))

(define (mntpoint-mntoptions obj val)
  (set-object-option!
   obj 'mntoptions
   (and val
        (or (assoc-ref well-known-mountpoints val)
            (cond ((isntfs? obj)
                   (string-append "umask=0"
                                  (let ((lang (getenv "LANG")))
                                    (if lang (string-append ",locale=" lang) ""))))
                  ((isfat? obj) "umask=0,quiet,showexec,iocharset=utf8,codepage=866")
                  (#t "nosuid,nodev,noexec"))))))

(define (mntpoint-forcemkfs obj val)
  (set-object-option!
   obj 'force-mkfs
   (and val (string? val)
        (not (procedure-property obj 'new))
        (not (or (string=? "/opt" val) (string-prefix? "/opt/" val)
                 (string=? "/mnt" val) (string-prefix? "/mnt/" val)
                 (string=? "/home" val) (string-prefix? "/home/" val)
                 (string=? "/usr/local" val) (string-prefix? "/usr/local/" val)
                 (string=? "/boot/efi" val) (string-prefix? "/boot/" val)))
        (and (or (member val (map car well-known-mountpoints))
                 (any (lambda (dir) (string-prefix? dir val)) '("/usr/" "/var/")))
             #t))))

(define (mntpoint< vola volb)
  (string< (mntpoint? vola) (mntpoint? volb)))

(define (assigned-volumes)
  (filter mntpoint?
          (filter (lambda(v) (not (isswap? v)))
                  (map object (evms-get-volume-list 0 0 #f)))))

(define (assigned-mountpoints)
  (map mntpoint? (assigned-volumes)))

(define (swap-volumes)
  (filter isswap? (map object (evms-get-volume-list 0 0 #f))))

(define (assigned+swap)
  (append (sort (assigned-volumes) mntpoint<)
          (swap-volumes)))

(define (maybe-mkfs vol)
  (or (procedure-property vol 'new)     ; already done by user
      (not (object-option-value vol 'force-mkfs))  ; keep as is
      (let ((fsim (assv-ref (vol) 'fsim))
            (fstype (evms-get-dev-type (assv-ref (vol) 'device))))
        (format #t "FORCE-MKFS: ~S ~S ~S\n" (assv-ref (vol) 'device) (assv-ref (fsim) 'name) fstype)
        (vol 'unmkfs)
        (let ((t (task 'mkfs fsim)))
          (set! (t 'selected) (list vol))
          (cond ((string=? "Ext2/3" (assv-ref (fsim) 'name))
                 (set! (t 'options) `(("journal" . ,(string=? "ext3" fstype)))))
                (#t #t))
          (t 'apply)
          (evms-commit-changes)))))

(define (guess-compat-device vol)
  (let ((device (assv-ref (vol) 'device)))
    (and (memq 'compatibility (assv-ref (vol) 'flags))
         (string-append
          "/dev/"
          (cond ((string-prefix? "/dev/evms/lvm2/" device)
                 (substring device 15))
                ((string-prefix? "/dev/evms/md/" device)
                 (substring device 13))
                (#t (substring device 10)))))))

(define (addmntent obj port)
  (let ((device (assv-ref (obj) 'device))
        (compat-device (guess-compat-device obj)))
    (format #t "ADDMNTENT: ~S compat: ~S\n" device compat-device)
    (format
     port
     "~A\t~A\t~A\t~A\t~A\t~A\n"
     ;; uuid/label/device
     (or (and-let* ((uuid (or (evms-get-dev-uuid compat-device) (evms-get-dev-uuid device)))) (string-append "UUID=" uuid))
         (and-let* ((label (or (evms-get-dev-label compat-device) (evms-get-dev-label device)))) (string-append "LABEL=" label))
         (or compat-device device))
     ;; mountpoint
     (if (isswap? obj) "swap" (object-option-value obj 'mntpoint))
     ;; fstype
     (cond ((isswap? obj) "swap")
           ((isntfs? obj) "ntfs-3g")
           (#t (or (evms-get-dev-type compat-device)
                   (evms-get-dev-type device)
                   "auto")))
     ;; mount options
     (or (object-option-value obj 'mntoptions)
         (assv-ref well-known-mountpoints (object-option-value obj 'mntpoint))
         (and (isswap? obj) "defaults")
         "nosuid,nodev,noexec")
     ;; fs_freq
     (if (isswap? obj) "0" (object-option-value obj 'mntfreq))
     ;; fs_passno
     (cond ((isswap? obj) "0")
           ((string-prefix? "/mnt/" (object-option-value obj 'mntpoint)) "0")
           ((and (not (string=? "/" (object-option-value obj 'mntpoint)))
                 (eqv? (object-option-value obj 'mntpass) 1)) "2")
           (#t (object-option-value obj 'mntpass))))))

(define (mountvol vol)             ; mount with defaults to prevent RO
  (if (isswap? vol) (evms-swapon (assv-ref (vol) 'device))
      (and-let* ((mntpoint (object-option-value vol 'mntpoint)))
                (or (memv 'not-mountable (assv-ref (vol) 'flags))
                    (vol 'mount (string-append changed-root mntpoint) "")))))

(define storage-id
  (let ((ids '()))
    (lambda (in)
      (case (storage-object-info-t-object-type in)
        ((disk)
         (let ((name (storage-object-info-t-sname in)))
           (if (pair? (assoc name ids)) (assoc-ref ids name)
               (let ((id (and-let*
                          ((port (false-if-exception
                                  (open-input-pipe
                                   (cond ((string-prefix? "sd" name)
                                          (string-append "cat /sys/block/" name "/device/model"))
                                         ((string-prefix? "hd" name)
                                          (string-append "cat /proc/ide/" name "/model"))
                                         (#t "cat /dev/null")))))
                           (id (false-if-exception (read-line port 'trim))))
                          (close-pipe port)
                          (if (eof-object? id) #f
                              id))))
                 (set! ids (acons name id ids))
                 id))))
        (else #f)))))

(define (reread-ptables)
  (for-each
   (lambda (disk)
     (let* ((name (string-join (cons "/dev" (string-split (assv-ref (disk) 'name) #\|)) "/" 'infix)))
       (format #t "BLKRRPART: ~S\n" name)
       (or (false-if-exception (begin (evms-reread-ptable name) #t))
           (throw 'swig-system-error (string-append "Unable to reread partition table on " name)
                  " reboot, then continue installation."))))
   (map object (evms-get-object-list 'disk #f 0 0 '(not-top)))))

;;----------------------------------------------------------------------
;;          object operations
;;----------------------------------------------------------------------
(define object-operations
  (let ((iops
         `((#t
            . ((type . ,evms-get-handle-object-type)
               (delete? . ,evms-can-delete) (delete . ,evms-delete)
               (destroy? . ,evms-can-destroy) (destroy . ,evms-destroy)
               (expand? . ,evms-can-expand) (expand . ,evms-expand)
               (shrink? . ,evms-can-shrink) (shrink . ,evms-shrink)
               (replace? . ,evms-can-replace) (replace . ,evms-replace)
               (set-info? . ,evms-can-set-info) (set-info . ,evms-set-info)
               (assign? . ,(lambda x #t)) (assign . ,evms-assign) ;; just for symmetry
               (unassign? . ,evms-can-unassign) (unassign . ,evms-unassign)
               (expand-points . ,evms-get-expand-points)
               (shrink-points . ,evms-get-shrink-points)
               (features . ,(lambda (h) (map plugin (evms-get-feature-list h))))))
           (volume
            . ((create-volume? . ,evms-can-create-volume)
               (create-volume . ,evms-create-volume)
               (create-compat? . ,evms-can-create-compatibility-volume)
               (create-compat . ,evms-create-compatibility-volume)))))
        (eops
         `((volume
            . ((set-name? . ,evms-can-set-volume-name)
               (set-name . ,evms-set-volume-name)
               (convert? . ,evms-can-convert-to-evms-volume)
               (convert . ,evms-convert-to-evms-volume)
               (convert-compat? . ,evms-can-convert-to-compatibility-volume)
               (convert-compat . ,evms-convert-to-compatibility-volume)
               (add-feature? . ,evms-can-add-feature-to-volume)
               (add-feature . ,evms-add-feature-to-volume)
               (mkfs? . ,evms-can-mkfs)
               (mkfs . ,evms-mkfs)
               (unmkfs? . ,evms-can-unmkfs)
               (unmkfs . ,evms-unmkfs)
               (fsck? . ,evms-can-fsck)
               (fsck . ,evms-fsck)
               (mount? . ,evms-can-mount)
               (mount . ,evms-mount)
               (umount? . ,evms-can-unmount)
               (umount . ,evms-unmount)
               (remount? . ,evms-can-remount)
               (remount . ,evms-remount))))))
    (map
     (lambda (x)
       (cons x
             (apply append
                    (or (assq-ref eops x) '())
                    (map cdr (assq-remove! iops x)))))
     '(disk segment region feature container volume))))

;;----------------------------------------------------------------------
;;          object properties
;;----------------------------------------------------------------------
(define object-properties
  (let ((props
         `((plugin
            . ((name . ,plugin-info-t-short-name)
               (handle . ,plugin-info-t-handle)
               (id . ,plugin-info-t-id)
               (long-name . ,plugin-info-t-long-name)
               (oem-name . ,plugin-info-t-oem-name)
               (container . ,plugin-info-t-supports-containers)
               (version . ,plugin-info-t-vers)))
           (container
            . ((name . ,storage-container-info-t-sname)
               (group . ,(lambda (h) (object (storage-container-info-t-disk-group h))))
               (flags . ,storage-container-info-t-scflags)
               (handle . ,storage-container-info-t-handle)
               (consumed . ,(lambda (h) (map object (storage-container-info-t-consumed h))))
               (produced . ,(lambda (h) (map object (storage-container-info-t-produced h))))
               (plugin . ,(lambda (h) (plugin (storage-container-info-t-plugin h))))
               (size . ,storage-container-info-t-size)))
           (volume
            . ((name . ,logical-volume-info-t-sname)
               (device . ,logical-volume-info-t-snode)
               (mountpoint . ,logical-volume-info-t-mount-point)
               (major . ,logical-volume-info-t-dev-major)
               (minor . ,logical-volume-info-t-dev-minor)
               (fsim . ,(lambda (h) (plugin (logical-volume-info-t-file-system-manager h))))
               (group . ,(lambda (h) (object (logical-volume-info-t-disk-group h))))
               (flags . ,logical-volume-info-t-volflags)
               (fssize . ,logical-volume-info-t-fs-size)
               (minfssize . ,logical-volume-info-t-min-fs-size)
               (maxfssize . ,logical-volume-info-t-max-fs-size)
               (volsize . ,logical-volume-info-t-vol-size)
               (maxvolsize . ,logical-volume-info-t-max-vol-size)
               (object . ,(lambda (h) (object (logical-volume-info-t-object h))))
               (handle . ,logical-volume-info-t-handle)
               (serial . ,logical-volume-info-t-serial-number)))
           (#t
            . ((id . ,storage-id)
               (name . ,storage-object-info-t-sname)
               (handle . ,storage-object-info-t-handle)
               (plugin . ,(lambda (h) (plugin (storage-object-info-t-plugin h))))
               (type . ,storage-object-info-t-object-type)
               (major . ,storage-object-info-t-dev-major)
               (minor . ,storage-object-info-t-dev-minor)
               (start . ,storage-object-info-t-start)
               (size . ,storage-object-info-t-size)
               (data-type . ,storage-object-info-t-data-type)
               (flags . ,storage-object-info-t-soflags)
               (group . ,(lambda (h) (object (storage-object-info-t-disk-group h))))
               (volume . ,(lambda (h) (object (storage-object-info-t-volume h))))
               (consumer . ,(lambda (h) (object (storage-object-info-t-consuming-container h))))
               (producer . ,(lambda (h) (object (storage-object-info-t-producing-container h))))
               (parents . ,(lambda (handle) (map object (storage-object-info-t-parents handle))))
               (children . ,(lambda (handle) (map object (storage-object-info-t-children handle))))
               (geometry . ,(lambda (info)
                              (mapkey (storage-object-info-t-geometry info)
                                      `((cylinders . ,geometry-t-cylinders)
                                        (heads . ,geometry-t-heads)
                                        (sectors/track . ,geometry-t-sectors-per-track)
                                        (bytes/sector . ,geometry-t-bytes-per-sector)
                                        (boot-cylinder-limit . ,geometry-t-boot-cylinder-limit)
                                        (block-size . ,geometry-t-block-size))))))))))
    (map
     (lambda (type)
       (cons type
             (assq-ref props
                       (or (not (memq type '(plugin container volume))) type))))
     '(plugin disk segment region feature container volume))))

;;----------------------------------------------------------------------
;;          object constructor
;;----------------------------------------------------------------------
(define object-list '())

(define (object handle)
  (and (not (eqv? handle 0))
       (or (assv-ref object-list handle)
           (let* ((type (evms-get-handle-object-type handle))
                  (ops (assv-ref object-operations type))
                  (props (assv-ref object-properties type))
                  (info
                   (case type
                     ((container) handle-object-info-t-storage-container-info-t-get)
                     ((volume) handle-object-info-t-logical-volume-info-t-get)
                     (else handle-object-info-t-storage-object-info-t-get)))
                  (obj
                   (lambda x
                     (if (null? x)
                         (mapkey (info (new-handle-object-info-t handle)) props)
                         (let ((cmd (assv-ref ops (car x)))
                               (args (map any->handle (cdr x))))
                           (if (procedure? cmd)
                               (apply cmd handle args)
                               (error 'unknown-method (car x))))))))
             (set-procedure-property! obj 'type type)
             (set-procedure-property! obj 'handle handle)
             (clear-object-options! obj)
             (set! object-list (acons handle obj object-list))
             obj))))

;;------------------------------------------------------------------
;;          plugin constructor
;;------------------------------------------------------------------
(define plugin-list '())

(define fsims '("SWAPFS" "Ext2/3" "XFS" "JFS" "ReiserFS" "FAT16" "FAT32" "NTFS" "OGFS" "OCFS2" "BtrFS")) ; that sucks, still

(define (plugin handle)
  (and (not (eqv? handle 0))
       (or (assv-ref plugin-list handle)
           (let* ((props (assv-ref object-properties 'plugin))
                  (obj (lambda ()
                         (mapkey
                          (handle-object-info-t-plugin-info-t-get
                           (new-handle-object-info-t handle))
                          props))))
             (set-procedure-properties!
              obj
              `((type . ,(if (member (assv-ref (obj) 'name) fsims) 'fsim 'plugin))
                (handle . ,handle)
                (properties ,(map car props))))
             (set! plugin-list (acons handle obj plugin-list))
             obj))))

;;------------------------------------------------------------------
;;          task constructor
;;------------------------------------------------------------------
(define (task action obj . args)
  (let ((handle
         (apply evms-create-task
                (let* ((handle (obj->handle obj))
                       (type (evms-get-handle-object-type handle))
                       (act (or (and (eqv? type 'container)
                                     (case action
                                       ((set-info) 'set-container-info)
                                       ;; ((expand) 'expand-container) FIXME
                                       ;; ((shrink) 'shrink-container)
                                       (else #f))) action)))
                  (list handle act)))))
    (letrec
        ((map-option
          (let ((max (evms-get-option-count handle)))
            (lambda*
             (iter #:optional (n 0) (accum '()))
             (if (eq? n max) accum
                 (map-option
                  iter
                  (+ n 1)
                  (cons (iter (new-option-descriptor-ex-t handle n))
                        accum)))))))

      (let ((task
             (make-procedure-with-setter
              (lambda (command)
                (case command
                  ((apply)
                   (map object (evms-invoke-task handle)))
                  ((destroy)
                   (evms-destroy-task handle))
                  ((action)
                   (evms-get-task-action handle))
                  ((acceptable)
                   (map object (evms-get-acceptable-objects handle)))
                  ((options)
                   (map-option
                    (lambda (option)
                      (cons
                       (option-descriptor-ex-t-name-get option)
                       (mapkey
                        option
                        `((title . ,option-descriptor-ex-t-title-get)
                          (tip . ,option-descriptor-ex-t-tip-get)
                          (help . ,option-descriptor-ex-t-help-get)
                          (unit . ,option-descriptor-ex-t-unit-get)
                          (flags . ,option-descriptor-ex-t-flags-get)
                          (constraint . ,option-descriptor-ex-t-constraint-get)
                          (value . ,option-descriptor-ex-t-value-get)))))))
                  (else
                   (error 'unknown-method command))))

              (lambda (command args)
                (case command
                  ((selected)
                   (apply
                    (lambda (declined effect)
                      (values effect (map object (map car declined))))
                    (evms-set-selected-objects handle (map obj->handle args))))
                  ((options)
                   (delete #f
                           (map-option
                            (lambda (option)
                              (let ((value (assoc (option-descriptor-ex-t-name-get option) args)))
                                (and (pair? value)
                                     (option-descriptor-ex-t-value-set option (cdr value))))))))
                  (else
                   (error 'unknown-method command)))))))
        (set-procedure-property! task 'context args)
        task))))

;;------------------------------------------------------------------
;;          task-like helpers for some storage ops
;;------------------------------------------------------------------
(define (utask action obj . args)
  (let ((task
         (make-procedure-with-setter
          (lambda (command)
            (case command
              ((action) action)
              ((acceptable) (list obj))
              ((apply)
               (case action
                 ((setmntent)
                  (apply-object-options! obj)
                  '())
                 ((mount)
                  (obj action
                       (object-option-value obj 'mntpoint)
                       (string-append "-o " (object-option-value obj 'mntoptions)))
                  '())
                 ((remount)
                  (obj action
                       (string-append "-o remount," (object-option-value obj 'mntoptions)))
                  '())
                 ((unmkfs)
                  (clear-object-options! obj)
                  (set-procedure-property! obj 'new #t)
                  (obj action)
                  (list obj))
                 ((create-volume create-compat)  ; return created volume
                  (obj action)
                  (let ((vol (assv-ref (obj) 'volume)))
                    (set-procedure-property! vol 'new #t)
                    (list vol)))
                 (else (obj action) '())))
              ((options)
               (case action
                 ((setmntent)
                  (cond
                   ((isswap? obj)
                    (object-options obj 'mntoptions (if (procedure-property obj 'new) #f 'force-mkfs)))
                   (#t
                    (let ((assigned (assigned-mountpoints)))
                      (let-values (((first rest)
                                    (partition (lambda (mnt) (member mnt assigned))
                                               (delete-duplicates (append assigned (map car well-known-mountpoints))))))
                        (or (object-option-value obj 'mntpoint)
                            (set-object-option!
                             obj 'mntpoint
                             (cond ((isfat? obj)
                                    "/boot/efi")
                                   ((isntfs? obj)
                                    "/mnt/disk")
                                   ((islvm? obj)
                                    (let ((sane-mountpoints (filter (lambda(s) (not (member s bare-mountpoints))) rest)))
                                      (if (null? sane-mountpoints) ""
                                          (car sane-mountpoints))))
                                   (#t (car rest)))))
                        (set-procedure-property!
                         obj 'mntpoint
                         (assv-set! (object-option obj 'mntpoint)
                                    'constraint `(string list ,(append rest first))))
                        (object-options obj 'mntpoint 'mntoptions 'mntpass 'mntfreq
                                        (if (procedure-property obj 'new) #f 'force-mkfs)))))))
                 ((mount)
                  (object-options obj 'mntpoint 'mntoptions))
                 ((remount)
                  (object-options obj 'mntoptions))
                 (else '())))
              ((destroy)
               (case action
                 ((setmntent)
                  (reset-object-options! obj)
                  '())
                 (else '())))
              (else (error 'unknown-method command))))
          (lambda (command args)
            (case command
              ((options)
               (set-object-options! obj args))
              (else '()))))))
    (set-procedure-property! task 'context args)
    task))

;;----------------------------------------------------------------------
;;          object sticky properties aka options
;;----------------------------------------------------------------------
(define object-option procedure-property)

(define (for-each-option obj iter)
  (for-each
   iter
   (case (procedure-property obj 'type)
     ((volume) volume-object-options)
     (else '()))))

(define (for-each-option-name obj iter)
  (for-each
   iter
   (case (procedure-property obj 'type)
     ((volume) (map car volume-object-options))
     (else '()))))

(define (clear-object-options! obj)
  (for-each-option
   obj
   (lambda (option)
     (set-procedure-property! obj (car option) (alist-copy (cdr option))))))

(define (apply-object-options! obj)
  (for-each-option-name
   obj
   (lambda (option)
     (set-procedure-property!
      obj option
      (alist-delete 'oldvalue (object-option obj option))))))

(define (reset-object-options! obj)
  (for-each-option-name
   obj
   (lambda (option)
     (and-let*
      ((opt (object-option obj option))
       (value (assv 'oldvalue opt)))
      (set-procedure-property!
       obj option
       (assv-set! (alist-delete 'oldvalue opt) 'value (cdr value)))))))

(define (object-option-value obj option)
  (and-let* ((opt (object-option obj option)))
            (assv-ref opt 'value)))

(define (object-option-flags obj option)
  (and-let* ((opt (object-option obj option)))
            (assv-ref opt 'flags)))

(define (copy-object-option option)
  (cons (car option) (alist-copy (cdr option))))

(define (object-options obj . args)
  (map
   copy-object-option
   (filter (lambda (option) (memv (car option) args))
           (procedure-properties obj))))

(define (set-object-option! obj option value)
  (format #t "SET-OBJECT-OPTION: ~S ~S ~S\n" (assv-ref (obj) 'name) option value)
  (and-let*
   ((opt (object-option obj option)))
   (let ((oldvalue (assv-ref opt 'value)))
     (if (equal? value oldvalue) 'pass
         (let ((guards (or (assv-ref opt 'guards) '()))
               (hooks (or (assv-ref opt 'hooks) '())))
           (and (every (lambda (f) (f obj value)) guards)  ; #f by guards
                (let ((flags (delv 'no-initial-value (assv-ref opt 'flags))))
                  (set-procedure-property!
                   obj option
                   (assv-set!
                    (assv-set!
                     (assv-set! opt 'flags flags)
                     'oldvalue oldvalue)
                    'value value))
                  (let ((ret (delq 'pass (map (lambda (f) (f obj value)) hooks))))
                    (format #t "RET: ~S: ~S\n" option ret)
                    (or (null? ret)
                        'options)))))))))

(define (set-object-options! obj args)
  (let ((effect (map (lambda (opt) (set-object-option! obj (car opt) (cdr opt))) args)))
    (cond ((not (every values effect))
           (clear-object-options! obj)
           '())
          ((memv 'options effect) '(options))
          (#t '()))))

(define volume-object-options
  `((mntpoint
     (title . "Mountpoint")
     (flags expandable important)       ; that sucks
     (constraint string list ,(map car well-known-mountpoints))
     (guards ,mntpoint-nonempty? ,mntpoint-unique? ,mntpoint-sane-pathname? ,mntpoint-valid-for-fstype? ,mntpoint-valid-for-storage?)
     (hooks ,mntpoint-mntoptions ,mntpoint-forcemkfs)
     (value . #f))
    (mntoptions
     (title . "Mount Options")
     (flags)
     (constraint string #f #f)
     (value . #f))
    (mntfreq
     (title . "Dump Frequency")
     (flags automatic not-required inactive)
     (constraint number list (0 1 2 3))
     (value . 1))
    (mntpass
     (title . "Check order")
     (flags automatic not-required inactive)
     (constraint number list (0 1 2 3))
     (value . 1))
    (force-mkfs
     (title . "Format before using")
     (flags automatic not-required no-initial-value)
     (constraint boolean #f #f)
     (value . #t))))

;;----------------------------------------------------------------------
;;          engine entry point
;;----------------------------------------------------------------------

(define* (engine #:key node (mode 'ro) callbacks (debug-level 'default) logfile prefix)
  (cond ((not prefix) #t)
        ((and (access? prefix R_OK)
              (eqv? (stat:type (stat prefix)) 'directory))
         (set! changed-root prefix)
         #t)
        (#t (throw 'swig-system-error prefix "inaccesible or not a directory")))

  (set! object-list '())
  (set! plugin-list '())
  (evms-open-engine node mode callbacks debug-level logfile)

  (lambda (command . args)
    (apply
     (case command
       ;; get plugin by name
       ((plugin)
        (lambda* (#:key name id)
                 (plugin
                  (or (and name (evms-get-plugin-by-name name))
                      (and id (evms-get-plugin-by-id id))
                      (error "no name nor id given")))))
       ;; list plugins
       ((plugin-list)
        (lambda* (#:key code flags)
                 ;; evms_plugin_code_t code => symbol: device segment region feature assoc fsim cluster dlm
                 ;; plugin_search_flags_t flags => symbol: containers
                 (map plugin (evms-get-plugin-list code flags))))
       ;; get object by name & type
       ((object)
        (lambda* (#:key type name)
                 (object (evms-get-object-handle-for-name type name))))
       ;; list objects
       ((object-list)
        (lambda* (#:key type data-type plugin group flags)
                 ;; object_type_t type => symbol: disk segment region feature container volume
                 ;; data_type_t data-type => symbol: meta data free
                 ;; plugin_handle_t plugin => <plugin>
                 ;; object_handle_t group => <object>
                 ;; object_search_flags_t flags => symbols: pending top not-top writeable no-disk-group input
                 (map object (evms-get-object-list
                              type data-type
                              (or (obj->handle plugin) 0)
                              (or (obj->handle group) 0) flags))))
       ;; list containers
       ((container-list)
        (lambda* (#:key plugin group flags)
                 ;; plugin_handle_t plugin => <plugin>
                 ;; object_handle_t group => <object>
                 ;; container_search_flags_t flags => symbols: pending no-disk-group
                 (map object (evms-get-container-list
                              (or (obj->handle plugin) 0)
                              (or (obj->handle group) 0) flags))))
       ;;((volume))
       ;; list volumes
       ((volume-list)
        (lambda* (#:key fsim group flags)
                 ;; object_handle_t fsim => <plugin>
                 ;; object_handle_t group => <object>
                 ;; volume_search_flags_t flags => symbols: pending no-disk-group
                 (map object (evms-get-volume-list
                              (or (obj->handle fsim) 0)
                              (or (obj->handle group) 0) flags))))
       ;; task
       ;; FIXME: message add-feature plugin-function
       ((create create-container assign set-info expand shrink slide move mkfs fsck)
        (lambda x (apply task command x)))

       ;; task-like iface
       ((create-volume create-compat unassign unmkfs setmntent mount umount remount delete destroy)
        (lambda x (apply utask command x)))

       ;; FIXME create object(s) from object(s)
       ;;         ((create)
       ;;          (lambda (x)
       ;;            (map object (apply evms-create x))))
       ;; FIXME create container from given objs
       ;;         ((container)
       ;;          (lambda (x)
       ;;            (object (evms-create-container x))))

       ((pending) evms-changes-pending)
       ((options)
        (lambda (obj args)
          (map (lambda (option)
                 (cons (car option) (assv-ref (cdr option) 'value)))
               (apply object-options obj args))))
       ((assigned)
        (lambda ()
          (filter-map
           (lambda(obj) (object-option-value obj 'mntpoint))
           (map object (evms-get-volume-list 0 0 #f)))))
       ((postponed)
        (lambda ()
          (filter-map
           (lambda(obj) (and
                         (not (isswap? obj))
                         (object-option-value obj 'mntpoint)
                         (object-option-value obj 'force-mkfs)
                         (assv-ref (obj) 'name)))
           (map object (evms-get-volume-list 0 0 #f)))))
       ((finalize)
        (lambda ()
          (and changed-root
               (let ((port (open-output-file "/tmp/fstab")))
                 (reread-ptables)
                 (for-each
                  (lambda (vol)
                    (maybe-mkfs vol)
                    (addmntent vol port)
                    (mountvol vol))
                  (assigned+swap))
                 (close-output-port port)))))
       ((commit) evms-commit-changes)
       ((version) evms-get-api-version)
       ((update) evms-update-evms-dev-tree)
       ((close) evms-close-engine)
       (else error)) args)))
